# vim: set filetype=python fileencoding=utf-8:
# -*- coding: utf-8 -*-

#============================================================================#
#                                                                            #
#  Licensed under the Apache License, Version 2.0 (the "License");           #
#  you may not use this file except in compliance with the License.          #
#  You may obtain a copy of the License at                                   #
#                                                                            #
#      http://www.apache.org/licenses/LICENSE-2.0                            #
#                                                                            #
#  Unless required by applicable law or agreed to in writing, software       #
#  distributed under the License is distributed on an "AS IS" BASIS,         #
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  #
#  See the License for the specific language governing permissions and       #
#  limitations under the License.                                            #
#                                                                            #
#============================================================================#


''' Rich annotations, public interfaces for cutomization functions, etc....

    .. note::

       By default, this module exports ``typing_extensions.Doc``, which is
       based on the withdrawn :pep:`727`. A ``typing_extensions`` maintainer,
       who was also the sponsor of this PEP, has indicated `in the PEP 727
       Discourse thread
       <https://discuss.python.org/t/pep-727-documentation-metadata-in-typing/32566/183>`_
       that ``typing_extensions`` will support ``Doc`` indefinitely. However,
       if it should disappear from ``typing_extensions``, we provide a
       compatible fallback. Unless you are using ``typing_extensions.Doc`` for
       other purposes, it is recommended that you import it from this package
       instead, to ensure future availability.
'''


from . import __
from . import nomina as _nomina


try: from typing_extensions import Doc # pyright: ignore[reportAssignmentType]
except ImportError: # pragma: no cover

    @__.dcls.dataclass( frozen = True, slots = True )
    class Doc:
        ''' Description of argument or attribute.

            Compatible with :pep:`727` ``Doc`` objects.
        '''

        documentation: str


Fragment: __.typx.TypeAlias = str | Doc
Fragments: __.typx.TypeAlias = __.cabc.Sequence[ Fragment ]


@__.dcls.dataclass( frozen = True, slots = True )
class Fname:
    ''' Name of documentation fragment in table. '''

    name: __.typx.Annotated[
        str, Doc( ''' Index to look up content in fragments table. ''' ) ]


@__.dcls.dataclass( frozen = True, slots = True )
class Raises:
    ''' Class and description of exception which can be raised.

        Should appear in the return annotations for a function.
    '''

    classes: __.typx.Annotated[
        type[ BaseException ] | __.cabc.Sequence[ type[ BaseException ] ],
        Doc( ''' Exception class or classes which can be raised. ''' ),
    ]
    description: __.typx.Annotated[
        __.typx.Optional[ str ],
        Doc( ''' When and why the exception is raised. ''' ),
    ] = None


AnnotationsArgument: __.typx.TypeAlias = __.typx.Annotated[
    _nomina.Annotations,
    Doc( ''' Annotations mapping for documentable object. ''' ),
]
FragmentsTableArgument: __.typx.TypeAlias = __.typx.Annotated[
    _nomina.FragmentsTable,
    Doc( ''' Table from which to copy docstring fragments. ''' ),
]
GlobalsLevelArgument: __.typx.TypeAlias = __.typx.Annotated[
    int,
    Doc(
        ''' Stack frame level from which to obtain globals.

            Default is 2, which is the caller of the caller.
        ''' ),
]
NotifierLevelArgument: __.typx.TypeAlias = __.typx.Annotated[
    _nomina.NotificationLevels,
    Doc( ''' Severity level of the notification. ''' ),
]
NotifierMessageArgument: __.typx.TypeAlias = __.typx.Annotated[
    str,
    Doc( ''' Message content to notify about. ''' ),
]
PossessorArgument: __.typx.TypeAlias = __.typx.Annotated[
    _nomina.Documentable,
    Doc(
        ''' Object being documented.

            May be a module, class, or function.
        ''' ),
]
PossessorClassArgument: __.typx.TypeAlias = __.typx.Annotated[
    type, Doc( ''' Class being documented. ''' ) ]
PossessorFunctionArgument: __.typx.TypeAlias = __.typx.Annotated[
    __.cabc.Callable[ ..., __.typx.Any ],
    Doc( ''' Function being documented. ''' ),
]
PossessorModuleArgument: __.typx.TypeAlias = __.typx.Annotated[
    __.types.ModuleType, Doc( ''' Module being documented. ''' ) ]
VisibilityAnnotationArgument: __.typx.TypeAlias = __.typx.Annotated[
    __.typx.Any, Doc( ''' Type annotation of the attribute. ''' ) ]
VisibilityDescriptionArgument: __.typx.TypeAlias = __.typx.Annotated[
    __.typx.Optional[ str ],
    Doc( ''' Optional description text for the attribute. ''' ),
]
VisibilityNameArgument: __.typx.TypeAlias = __.typx.Annotated[
    str, Doc( ''' Name of the attribute being evaluated. ''' ) ]


class Sentinels( __.enum.Enum ):
    ''' Sentinel values used in various parts of the package. '''

    Absent      = __.enum.auto( )
    Incomplete  = __.enum.auto( )


absent: __.typx.Annotated[
    Sentinels, Doc( ''' Indicates annotation or other data is missing. ''' )
] = Sentinels.Absent
incomplete: __.typx.Annotated[
    Sentinels, Doc( ''' Indicates annotation reduction is incomplete. ''' ),
] = Sentinels.Incomplete


@__.dcls.dataclass( frozen = True, kw_only = True, slots = True )
class AdjunctsData:
    ''' Data about type-adjacent entities. '''

    extras: __.typx.Annotated[
        __.cabc.MutableSequence[ __.typx.Any ],
        Doc( ''' Additional annotations. ''' ),
    ] = __.dcls.field( default_factory = list[ __.typx.Any ] )
    traits: __.typx.Annotated[
        __.cabc.MutableSet[ str ],
        Doc( ''' Trait names collected during annotation processing. ''' ),
    ] = __.dcls.field( default_factory = set[ str ] )

    def copy( self ) -> __.typx.Self:
        ''' Creates a shallow copy of the adjuncts data. '''
        return type( self )(
            extras = list( self.extras ),
            traits = set( self.traits ) )


@__.dcls.dataclass( frozen = True, kw_only = True, slots = True )
class AnnotationsCache:
    ''' Lookup table for reduced annotations from original annotations.

        Has special values for absent and incomplete entries.
    '''

    entries: __.typx.Annotated[
        dict[ __.typx.Any, __.typx.Any ],
        Doc( ''' Mapping from original annotations to reduced forms. ''' ),
    ] = __.dcls.field( default_factory = dict[ __.typx.Any, __.typx.Any ] )

    def access(
        self, original: __.typx.Annotated[
            __.typx.Any,
            Doc( ''' Original annotation to look up in cache. ''' ),
        ]
    ) -> __.typx.Annotated[
        __.typx.Any,
        Doc(
            ''' Reduced annotation from cache.

                Absence sentinel if not found.
            ''' ),
    ]:
        ''' Accesses entry value, if it exists. '''
        try: return self.entries.get( original, absent )
        except TypeError: return self.entries.get( id( original ), absent )

    def enter(
        self,
        original: __.typx.Annotated[
            __.typx.Any,
            Doc( ''' Original annotation to use as cache key. ''' ),
        ],
        reduction: __.typx.Annotated[
            __.typx.Any,
            Doc( ''' Reduced form of annotation to store as value. ''' ),
        ] = incomplete,
    ) -> __.typx.Any:
        ''' Adds reduced annotation to cache, returning it.

            Cache key is original annotation.
            If reduction is not specified, then an incompletion sentinel is
            added as the value for the entry.
        '''
        try: self.entries[ original ] = reduction
        except TypeError: self.entries[ id( original ) ] = reduction
        return reduction


class AttributeAssociations( __.enum.Enum ):
    ''' Association level of an attribute with its containing entity. '''

    Module      = __.enum.auto( )
    Class       = __.enum.auto( )
    Instance    = __.enum.auto( )


class ValuationModes( __.enum.Enum ):
    ''' Annotation for how default value is determined.

        Accept means to use assigned value.
        Suppress means to use no value.
        Surrogate means to use surrogate value.
    '''

    Accept      = __.enum.auto( )
    Suppress    = __.enum.auto( )
    Surrogate   = __.enum.auto( )


@__.dcls.dataclass( frozen = True, kw_only = True, slots = True )
class Default:
    ''' How to process default value. '''

    mode: __.typx.Annotated[
        ValuationModes,
        Doc( ''' Method for handling default value processing. ''' ),
    ] = ValuationModes.Accept
    surrogate: __.typx.Annotated[
        __.typx.Any,
        Doc(
            ''' Alternative value to use when surrogate mode.

                Usually a description string.
            ''' ),
    ] = absent


class FragmentSources( __.enum.Enum ):
    ''' Possible sources for documentation fragments. '''

    Annotation  = __.enum.auto( )
    Argument    = __.enum.auto( ) # *fragments
    Attribute   = __.enum.auto( ) # _dynadoc_fragments_
    Docstring   = __.enum.auto( )
    Renderer    = __.enum.auto( )


class FragmentRectifier( __.typx.Protocol ):
    ''' Cleans and normalizes documentation fragment. '''

    @staticmethod
    def __call__(
        fragment: __.typx.Annotated[
            str,
            Doc( ''' Raw fragment text to be cleaned and normalized. ''' ),
        ],
        source: __.typx.Annotated[
            FragmentSources,
            Doc(
                ''' Source type of fragment for context-aware processing.
                ''' ),
        ],
    ) -> str:
        ''' (Signature for fragment rectifier.) '''
        raise NotImplementedError # pragma: no cover


@__.dcls.dataclass( frozen = True, kw_only = True, slots = True )
class InformationBase:
    ''' Base for information on various kinds of entities. '''

    annotation: __.typx.Annotated[
        __.typx.Any,
        Doc( ''' Type annotation associated with this entity. ''' ),
    ]
    description: __.typx.Annotated[
        __.typx.Optional[ str ],
        Doc( ''' Human-readable description of the entity. ''' ),
    ]


@__.dcls.dataclass( frozen = True, kw_only = True, slots = True )
class ArgumentInformation( InformationBase ):
    ''' Information about a function argument. '''

    name: __.typx.Annotated[
        str,
        Doc( ''' Name of the function parameter. ''' ),
    ]
    paramspec: __.typx.Annotated[
        __.inspect.Parameter,
        Doc( ''' Inspection parameter object with various details. ''' ),
    ]
    default: __.typx.Annotated[
        Default,
        Doc( ''' Configuration for how to handle default value. ''' ),
    ]


@__.dcls.dataclass( frozen = True, kw_only = True, slots = True )
class AttributeInformation( InformationBase ):
    ''' Information about a class or module attribute. '''

    name: __.typx.Annotated[
        str,
        Doc( ''' Name of the attribute. ''' ),
    ]
    association: __.typx.Annotated[
        AttributeAssociations,
        Doc( ''' Attribute associated with module, class, or instance? ''' ),
    ]
    default: __.typx.Annotated[
        Default, Doc( ''' How to handle default value. ''' ) ]


@__.dcls.dataclass( frozen = True, kw_only = True, slots = True )
class ExceptionInformation( InformationBase ):
    ''' Information about an exception that can be raised. '''
    pass


@__.dcls.dataclass( frozen = True, kw_only = True, slots = True )
class ReturnInformation( InformationBase ):
    ''' Information about a function's return value. '''
    pass


Informations: __.typx.TypeAlias = __.cabc.Sequence[ InformationBase ]


class Notifier( __.typx.Protocol ):
    ''' Notifies of warnings and errors. '''

    @staticmethod
    def __call__(
        level: NotifierLevelArgument,
        message: NotifierMessageArgument,
    ) -> None:
        ''' (Signature for notifier callback.) '''
        raise NotImplementedError # pragma: no cover


class Visibilities( __.enum.Enum ):
    ''' Annotation to determine visibility of attribute.

        Default means to defer to visibility predicate in use.
        Conceal means to hide regardless of visibility predicate.
        Reveal means to show regardless of visibility predicate.
    '''

    Default     = __.enum.auto( )
    Conceal     = __.enum.auto( )
    Reveal      = __.enum.auto( )


class VisibilityDecider( __.typx.Protocol ):
    ''' Decides if attribute should have visible documentation. '''

    @staticmethod
    def __call__(
        possessor: PossessorArgument,
        name: VisibilityNameArgument,
        annotation: VisibilityAnnotationArgument,
        description: VisibilityDescriptionArgument,
    ) -> bool:
        ''' (Signature for visibility decider.) '''
        raise NotImplementedError # pragma: no cover


AnnotationsCacheArgument: __.typx.TypeAlias = __.typx.Annotated[
    AnnotationsCache,
    Doc(
        ''' Cache for storing reduced annotation forms.

            Also used for cycle detection.
        ''' ),
]
InformationsArgument: __.typx.TypeAlias = __.typx.Annotated[
    Informations,
    Doc(
        ''' Sequence of information blocks from object introspection.

            Information may be about arguments to a function, attributes on a
            class or module, exceptions raised by a function, or returns from a
            function.
        ''' ),
]



# vim: set filetype=python fileencoding=utf-8:
# -*- coding: utf-8 -*-

#============================================================================#
#                                                                            #
#  Licensed under the Apache License, Version 2.0 (the "License");           #
#  you may not use this file except in compliance with the License.          #
#  You may obtain a copy of the License at                                   #
#                                                                            #
#      http://www.apache.org/licenses/LICENSE-2.0                            #
#                                                                            #
#  Unless required by applicable law or agreed to in writing, software       #
#  distributed under the License is distributed on an "AS IS" BASIS,         #
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  #
#  See the License for the specific language governing permissions and       #
#  limitations under the License.                                            #
#                                                                            #
#============================================================================#


''' Docstring generation from annotations with configurable output formats. '''


from . import __
from . import assembly
from . import context
from . import factories
from . import interfaces
from . import introspection
from . import nomina
from . import renderers
from . import xtnsapi
# --- BEGIN: Injected by Copier ---
# --- END: Injected by Copier ---

from .userapi import *


__version__: __.typx.Annotated[ str, Visibilities.Reveal ]
__version__ = '1.2.1'


def _notify( level: NotificationLevels, message: str ) -> None:
    ''' Issues warning message. (Internal use within this package itself.) '''
    __.warnings.warn( # pyright: ignore[reportCallIssue]
        message, category = RuntimeWarning, stacklevel = 2 )


_context = produce_context( notifier = _notify )
_introspection_cc = ClassIntrospectionControl(
    inheritance = True,
    introspectors = ( introspection.introspect_special_classes, ) )
_introspection = IntrospectionControl(
    class_control = _introspection_cc, targets = IntrospectionTargetsOmni )
assign_module_docstring(
    __.package_name,
    context = _context,
    introspection = _introspection,
    table = __.fragments )
# TODO: Reclassify package modules as immutable and concealed.



# vim: set filetype=python fileencoding=utf-8:
# -*- coding: utf-8 -*-

#============================================================================#
#                                                                            #
#  Licensed under the Apache License, Version 2.0 (the "License");           #
#  you may not use this file except in compliance with the License.          #
#  You may obtain a copy of the License at                                   #
#                                                                            #
#      http://www.apache.org/licenses/LICENSE-2.0                            #
#                                                                            #
#  Unless required by applicable law or agreed to in writing, software       #
#  distributed under the License is distributed on an "AS IS" BASIS,         #
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  #
#  See the License for the specific language governing permissions and       #
#  limitations under the License.                                            #
#                                                                            #
#============================================================================#


''' Data transfer objects for execution context. '''


from . import __
from . import interfaces as _interfaces
from . import nomina as _nomina


fragments_name_default = '_dynadoc_fragments_'
introspection_limit_name_default = '_dynadoc_introspection_limit_'


@__.dcls.dataclass( frozen = True, kw_only = True, slots = True )
class Context:

    _dynadoc_fragments_: __.typx.ClassVar[
        _interfaces.Fragments ] = ( 'context', )

    notifier: __.typx.Annotated[
        _interfaces.Notifier,
        _interfaces.Fname( 'notifier' ),
    ]
    fragment_rectifier: __.typx.Annotated[
        _interfaces.FragmentRectifier,
        _interfaces.Fname( 'fragment rectifier' ),
    ]
    visibility_decider: __.typx.Annotated[
        _interfaces.VisibilityDecider,
        _interfaces.Fname( 'visibility decider' ),
    ]
    fragments_name: __.typx.Annotated[
        str,
        _interfaces.Fname( 'fragments name' ),
    ] = fragments_name_default
    introspection_limit_name: __.typx.Annotated[
        str,
        _interfaces.Fname( 'introspection limit name' ),
    ] = introspection_limit_name_default
    invoker_globals: __.typx.Annotated[
        __.typx.Optional[ _nomina.Variables ],
        _interfaces.Fname( 'invoker globals' ),
    ] = None
    resolver_globals: __.typx.Annotated[
        __.typx.Optional[ _nomina.Variables ],
        _interfaces.Fname( 'resolver globals' ),
    ] = None
    resolver_locals: __.typx.Annotated[
        __.typx.Optional[ _nomina.Variables ],
        _interfaces.Fname( 'resolver locals' ),
    ] = None

    def with_invoker_globals(
        self,
        level: _interfaces.GlobalsLevelArgument = 2
    ) -> __.typx.Self:
        ''' Returns new context with invoker globals from stack frame. '''
        iglobals = __.inspect.stack( )[ level ].frame.f_globals
        return type( self )(
            notifier = self.notifier,
            fragment_rectifier = self.fragment_rectifier,
            visibility_decider = self.visibility_decider,
            fragments_name = self.fragments_name,
            introspection_limit_name = self.introspection_limit_name,
            invoker_globals = iglobals,
            resolver_globals = self.resolver_globals,
            resolver_locals = self.resolver_locals )


ContextArgument: __.typx.TypeAlias = __.typx.Annotated[
    Context, _interfaces.Fname( 'context' ) ]
IntrospectionArgumentFref: __.typx.TypeAlias = __.typx.Annotated[
    'IntrospectionControl', _interfaces.Fname( 'introspection' ) ]


class ClassIntrospector( __.typx.Protocol ):
    ''' Custom introspector for class annotations and attributes. '''

    @staticmethod
    def __call__( # noqa: PLR0913
        possessor: _interfaces.PossessorClassArgument, /,
        context: ContextArgument,
        introspection: IntrospectionArgumentFref,
        annotations: _interfaces.AnnotationsArgument,
        cache: _interfaces.AnnotationsCacheArgument,
        table: _interfaces.FragmentsTableArgument,
    ) -> __.typx.Optional[ _interfaces.Informations ]:
        ''' Introspects class and returns information about its members. '''
        raise NotImplementedError # pragma: no cover

ClassIntrospectors: __.typx.TypeAlias = __.cabc.Sequence[ ClassIntrospector ]


@__.dcls.dataclass( frozen = True, kw_only = True, slots = True )
class ClassIntrospectionLimit:
    ''' Limits on class introspection behavior. '''

    avoid_inheritance: __.typx.Annotated[
        bool,
        _interfaces.Doc( ''' Avoid introspecting inherited members? ''' ),
    ] = False
    ignore_attributes: __.typx.Annotated[
        bool,
        _interfaces.Doc(
            ''' Ignore attributes not covered by annotations? ''' ),
    ] = False


@__.dcls.dataclass( frozen = True, kw_only = True, slots = True )
class ClassIntrospectionControl:
    ''' Controls on class introspection behavior. '''

    inheritance: __.typx.Annotated[
        bool, _interfaces.Doc( ''' Inherit annotations? ''' )
    ] = False
    introspectors: __.typx.Annotated[
        ClassIntrospectors,
        _interfaces.Doc( ''' Custom introspectors to apply. ''' ),
    ] = ( )
    scan_attributes: __.typx.Annotated[
        bool,
        _interfaces.Doc( ''' Scan attributes not covered by annotations? ''' ),
    ] = False

    def with_limit(
        self,
        limit: __.typx.Annotated[
            ClassIntrospectionLimit,
            _interfaces.Doc(
                ''' Limits to apply to this introspection control. ''' ),
        ]
    ) -> __.typx.Self:
        ''' Returns new control with applied limits. '''
        inheritance = self.inheritance and not limit.avoid_inheritance
        scan_attributes = self.scan_attributes and not limit.ignore_attributes
        return type( self )(
            inheritance = inheritance,
            introspectors = self.introspectors,
            scan_attributes = scan_attributes )


@__.dcls.dataclass( frozen = True, kw_only = True, slots = True )
class ModuleIntrospectionLimit:
    ''' Limits on module introspection behavior. '''

    ignore_attributes: __.typx.Annotated[
        bool,
        _interfaces.Doc(
            ''' Ignore attributes not covered by annotations? ''' ),
    ] = False


@__.dcls.dataclass( frozen = True, kw_only = True, slots = True )
class ModuleIntrospectionControl:
    ''' Controls on module introspection behavior. '''

    scan_attributes: __.typx.Annotated[
        bool,
        _interfaces.Doc( ''' Scan attributes not covered by annotations? ''' ),
    ] = False

    def with_limit(
        self,
        limit: __.typx.Annotated[
            ModuleIntrospectionLimit,
            _interfaces.Doc(
                ''' Limits to apply to this introspection control. ''' ),
        ]
    ) -> __.typx.Self:
        ''' Returns new control with applied limits. '''
        scan_attributes = self.scan_attributes and not limit.ignore_attributes
        return type( self )( scan_attributes = scan_attributes )


class IntrospectionLimiter( __.typx.Protocol ):
    ''' Can return modified introspection control for attribute. '''

    @staticmethod
    def __call__(
        objct: __.typx.Annotated[
            object,
            _interfaces.Doc(
                ''' Object being evaluated for introspection limits. ''' ),
        ],
        introspection: IntrospectionArgumentFref,
    ) -> 'IntrospectionControl':
        ''' Returns modified introspection control with limits applied. '''
        raise NotImplementedError # pragma: no cover

IntrospectionLimiters: __.typx.TypeAlias = __.typx.Annotated[
    __.cabc.Sequence[ IntrospectionLimiter ],
    _interfaces.Doc(
        ''' Functions which can apply limits to introspection control. ''' ),
]


class IntrospectionTargets( __.enum.IntFlag ):
    ''' Kinds of objects to recursively document. '''

    Null        = 0
    Class       = __.enum.auto( )
    Descriptor  = __.enum.auto( )
    Function    = __.enum.auto( )
    Module      = __.enum.auto( )


IntrospectionTargetsSansModule: __.typx.Annotated[
    IntrospectionTargets,
    _interfaces.Doc( ''' All introspection targets except modules. ''' ),
] = (   IntrospectionTargets.Class
    |   IntrospectionTargets.Descriptor
    |   IntrospectionTargets.Function )
IntrospectionTargetsOmni: __.typx.Annotated[
    IntrospectionTargets,
    _interfaces.Doc(
        ''' All available introspection targets including modules. ''' ),
] = IntrospectionTargetsSansModule | IntrospectionTargets.Module


@__.dcls.dataclass( frozen = True, kw_only = True, slots = True )
class IntrospectionLimit:
    ''' Limits on introspection behavior. '''

    disable: __.typx.Annotated[
        bool, _interfaces.Doc( ''' Disable introspection? ''' )
    ] = False
    class_limit: __.typx.Annotated[
        ClassIntrospectionLimit,
        _interfaces.Doc( ''' Limits specific to class introspection. ''' ),
    ] = ClassIntrospectionLimit( )
    module_limit: __.typx.Annotated[
        ModuleIntrospectionLimit,
        _interfaces.Doc( ''' Limits specific to module introspection. ''' ),
    ] = ModuleIntrospectionLimit( )
    targets_exclusions: __.typx.Annotated[
        IntrospectionTargets,
        _interfaces.Doc( ''' Target types to exclude from introspection. ''' ),
    ] = IntrospectionTargets.Null


@__.dcls.dataclass( frozen = True, kw_only = True, slots = True )
class IntrospectionControl:

    _dynadoc_fragments_ = ( 'introspection', )

    enable: __.typx.Annotated[
        bool,
        _interfaces.Doc( ''' Whether introspection is enabled at all. ''' ),
    ] = True
    class_control: __.typx.Annotated[
        ClassIntrospectionControl,
        _interfaces.Doc( ''' Controls specific to class introspection. ''' ),
    ] = ClassIntrospectionControl( )
    module_control: __.typx.Annotated[
        ModuleIntrospectionControl,
        _interfaces.Doc( ''' Controls specific to module introspection. ''' ),
    ] = ModuleIntrospectionControl( )
    limiters: __.typx.Annotated[
        IntrospectionLimiters,
        _interfaces.Doc(
            ''' Functions that can apply limits to introspection. ''' ),
    ] = ( )
    targets: __.typx.Annotated[
        IntrospectionTargets,
        _interfaces.Doc(
            ''' Which types of objects to recursively document. ''' ),
    ] = IntrospectionTargets.Null
    # TODO? Maximum depth.
    #       (Suggested by multiple LLMs; not convinced that it is needed.)

    def evaluate_limits_for(
        self,
        objct: __.typx.Annotated[
            object,
            _interfaces.Doc( ''' Object to evaluate limits for. ''' ),
        ]
    ) -> 'IntrospectionControl':
        ''' Determine which introspection limits apply to object. '''
        introspection_ = self
        for limiter in self.limiters:
            introspection_ = limiter( objct, introspection_ )
        return introspection_

    def with_limit(
        self,
        limit: __.typx.Annotated[
            IntrospectionLimit,
            _interfaces.Doc(
                ''' Limits to apply to this introspection control. ''' ),
        ]
    ) -> __.typx.Self:
        ''' Returns new control with applied limits. '''
        enable = self.enable and not limit.disable
        class_control = self.class_control.with_limit( limit.class_limit )
        module_control = self.module_control.with_limit( limit.module_limit )
        targets = self.targets & ~limit.targets_exclusions
        return type( self )(
            enable = enable,
            class_control = class_control,
            module_control = module_control,
            limiters = self.limiters,
            targets = targets )


IntrospectionArgument: __.typx.TypeAlias = __.typx.Annotated[
    IntrospectionControl, _interfaces.Fname( 'introspection' ) ]



# vim: set filetype=python fileencoding=utf-8:
# -*- coding: utf-8 -*-

#============================================================================#
#                                                                            #
#  Licensed under the Apache License, Version 2.0 (the "License");           #
#  you may not use this file except in compliance with the License.          #
#  You may obtain a copy of the License at                                   #
#                                                                            #
#      http://www.apache.org/licenses/LICENSE-2.0                            #
#                                                                            #
#  Unless required by applicable law or agreed to in writing, software       #
#  distributed under the License is distributed on an "AS IS" BASIS,         #
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  #
#  See the License for the specific language governing permissions and       #
#  limitations under the License.                                            #
#                                                                            #
#============================================================================#


''' Factories and registries. '''
# TODO? Registry for deferred decoration.


from . import __
from . import xtnsapi as _xtnsapi


# _package_location = __.Path( __file__ ).parent
def notify( level: _xtnsapi.NotificationLevels, message: str ) -> None:
    # TODO: Python 3.12: Use 'skip_file_prefixes' option.
    ''' Issues warning message. '''
    __.warnings.warn( message, category = RuntimeWarning, stacklevel = 2 )
        # skip_file_prefixes = ( str( _package_location ), ) )


def rectify_fragment(
    fragment: str, source: _xtnsapi.FragmentSources
) -> str:
    ''' Cleans and normalizes fragment according to source. '''
    match source:
        case _xtnsapi.FragmentSources.Renderer: return fragment.strip( )
        case _: return __.inspect.cleandoc( fragment ).rstrip( )


def produce_context( # noqa: PLR0913
    invoker_globals: _xtnsapi.InvokerGlobalsArgument = None,
    resolver_globals: _xtnsapi.ResolverGlobalsArgument = None,
    resolver_locals: _xtnsapi.ResolverLocalsArgument = None,
    notifier: _xtnsapi.NotifierArgument = notify,
    fragment_rectifier: _xtnsapi.FragmentRectifierArgument = (
        rectify_fragment ),
    visibility_decider: _xtnsapi.VisibilityDeciderArgument = (
        _xtnsapi.is_attribute_visible ),
    fragments_name: _xtnsapi.FragmentsNameArgument = (
        _xtnsapi.fragments_name_default ),
    introspection_limit_name: _xtnsapi.IntrospectionLimitNameArgument = (
        _xtnsapi.introspection_limit_name_default ),
) -> _xtnsapi.Context:
    ''' Produces context data transfer object.

        Reasonable defaults are used for arguments that are not supplied.
    '''
    return _xtnsapi.Context(
        notifier = notifier,
        fragment_rectifier = fragment_rectifier,
        visibility_decider = visibility_decider,
        invoker_globals = invoker_globals,
        resolver_globals = resolver_globals,
        resolver_locals = resolver_locals )



# vim: set filetype=python fileencoding=utf-8:
# -*- coding: utf-8 -*-

#============================================================================#
#                                                                            #
#  Licensed under the Apache License, Version 2.0 (the "License");           #
#  you may not use this file except in compliance with the License.          #
#  You may obtain a copy of the License at                                   #
#                                                                            #
#      http://www.apache.org/licenses/LICENSE-2.0                            #
#                                                                            #
#  Unless required by applicable law or agreed to in writing, software       #
#  distributed under the License is distributed on an "AS IS" BASIS,         #
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  #
#  See the License for the specific language governing permissions and       #
#  limitations under the License.                                            #
#                                                                            #
#============================================================================#


''' Sphinx Autodoc reStructuredText renderers. '''


from . import __


class Style( __.enum.Enum ):
    ''' Style of formatter output. '''

    Legible     = __.enum.auto( )
    Pep8        = __.enum.auto( )


StyleArgument: __.typx.TypeAlias = __.typx.Annotated[
    Style,
    __.Doc(
        ''' Output style for renderer.

            Legible: Extra space padding inside of delimiters.
            Pep8: As the name implies.
        ''' ),
]


def produce_fragment(
    possessor: __.PossessorArgument,
    informations: __.InformationsArgument,
    context: __.ContextArgument,
    style: StyleArgument = Style.Legible,
) -> __.RendererReturnValue:
    ''' Produces a reStructuredText docstring fragment.

        Combines information from object introspection into a formatted
        docstring fragment suitable for Sphinx Autodoc.
    '''
    return '\n'.join(
        _produce_fragment_partial( possessor, information, context, style )
        for information in informations )


_qualident_regex = __.re.compile( r'''^([\w\.]+).*$''' )
def _extract_qualident( name: str, context: __.Context ) -> str:
    ''' Extracts a qualified identifier from a string representation.

        Used to extract the qualified name of an object from its string
        representation when direct name access is not available.
    '''
    extract = _qualident_regex.match( name )
    if extract is not None: return extract[ 1 ]
    return '<unknown>'


def _format_annotation(
    annotation: __.typx.Any, context: __.Context, style: Style
) -> str:
    ''' Formats a type annotation as a string for documentation.

        Handles various annotation types including unions, generics,
        and literals. Formats according to the selected style.
    '''
    if isinstance( annotation, list ):
        seqstr = ', '.join(
            _format_annotation( element, context, style )
            for element in annotation ) # pyright: ignore[reportUnknownVariableType]
        return _stylize_delimiter( style, '[]', seqstr )
    origin = __.typx.get_origin( annotation )
    if origin is None:
        return _qualify_object_name( annotation, context )
    arguments = __.typx.get_args( annotation )
    if origin in ( __.types.UnionType, __.typx.Union ):
        return ' | '.join(
            _format_annotation( argument, context, style )
            for argument in arguments )
    oname = _qualify_object_name( origin, context )
    if not arguments: return oname
    if origin is __.typx.Literal:
        argstr = ', '.join( repr( argument ) for argument in arguments )
    else:
        argstr = ', '.join(
            _format_annotation( argument, context, style )
            for argument in arguments )
    return _stylize_delimiter( style, '[]', argstr, oname )


def _format_description( description: __.typx.Optional[ str ] ) -> str:
    ''' Ensures that multiline descriptions render correctly. '''
    if not description: return ''
    lines = description.split( '\n' )
    lines[ 1 : ] = [ f"    {line}" for line in lines[ 1 : ] ]
    return '\n'.join( lines )


def _produce_fragment_partial(
    possessor: __.Documentable,
    information: __.InformationBase,
    context: __.Context,
    style: Style,
) -> str:
    ''' Produces a docstring fragment for a single piece of information.

        Dispatches to appropriate producer based on the type of information.
    '''
    if isinstance( information, __.ArgumentInformation ):
        return (
            _produce_argument_text( possessor, information, context, style ) )
    if isinstance( information, __.AttributeInformation ):
        return (
            _produce_attribute_text( possessor, information, context, style ) )
    if isinstance( information, __.ExceptionInformation ):
        return (
            _produce_exception_text( possessor, information, context, style ) )
    if isinstance( information, __.ReturnInformation ):
        return (
            _produce_return_text( possessor, information, context, style ) )
    context.notifier(
        'admonition', f"Unrecognized information: {information!r}" )
    return ''


def _produce_argument_text(
    possessor: __.Documentable,
    information: __.ArgumentInformation,
    context: __.Context,
    style: Style,
) -> str:
    ''' Produces reStructuredText for argument information.

        Formats function arguments in Sphinx-compatible reST format,
        including parameter descriptions and types.
    '''
    annotation = information.annotation
    description = _format_description( information.description )
    name = information.name
    lines: list[ str ] = [ ]
    lines.append(
        f":argument {name}: {description}"
        if description else f":argument {name}:" )
    if annotation is not __.absent:
        typetext = _format_annotation( annotation, context, style )
        lines.append( f":type {information.name}: {typetext}" )
    return '\n'.join( lines )


def _produce_attribute_text(
    possessor: __.Documentable,
    information: __.AttributeInformation,
    context: __.Context,
    style: Style,
) -> str:
    ''' Produces reStructuredText for attribute information.

        Formats class and instance attributes in Sphinx-compatible reST format.
        Delegates to special handler for module attributes.
    '''
    annotation = information.annotation
    match information.association:
        case __.AttributeAssociations.Module:
            return _produce_module_attribute_text(
                possessor, information, context, style )
        case __.AttributeAssociations.Class: vlabel = 'cvar'
        case __.AttributeAssociations.Instance: vlabel = 'ivar'
    description = _format_description( information.description )
    name = information.name
    lines: list[ str ] = [ ]
    lines.append(
        f":{vlabel} {name}: {description}"
        if description else f":{vlabel} {name}:" )
    if annotation is not __.absent:
        typetext = _format_annotation( annotation, context, style )
        lines.append( f":vartype {name}: {typetext}" )
    return '\n'.join( lines )


def _produce_module_attribute_text(
    possessor: __.Documentable,
    information: __.AttributeInformation,
    context: __.Context,
    style: Style,
) -> str:
    ''' Produces reStructuredText for module attribute information.

        Formats module attributes in Sphinx-compatible reST format,
        with special handling for TypeAlias attributes.
    '''
    annotation = information.annotation
    description = information.description or ''
    name = information.name
    match information.default.mode:
        case __.ValuationModes.Accept:
            value = getattr( possessor, name, __.absent )
        case __.ValuationModes.Suppress:
            value = __.absent
        case __.ValuationModes.Surrogate: # pragma: no branch
            value = __.absent
    lines: list[ str ] = [ ]
    if annotation is __.typx.TypeAlias:
        lines.append( f".. py:type:: {name}" )
        if value is not __.absent: # pragma: no branch
            value_ar = __.reduce_annotation(
                value, context,
                __.AdjunctsData( ),
                __.AnnotationsCache( ) )
            value_s = _format_annotation( value_ar, context, style )
            lines.append( f"   :canonical: {value_s}" )
        if description: lines.extend( [ '', f"   {description}" ] )
    else:
        # Note: No way to inject data docstring as of 2025-05-11.
        #       Autodoc will read doc comments and pseudo-docstrings,
        #       but we have no means of supplying description via a field.
        lines.append( f".. py:data:: {name}" )
        if annotation is not __.absent:
            typetext = _format_annotation( annotation, context, style )
            lines.append( f"    :type: {typetext}" )
        if value is not __.absent:
            lines.append( f"    :value: {value!r}" )
    return '\n'.join( lines )


def _produce_exception_text(
    possessor: __.Documentable,
    information: __.ExceptionInformation,
    context: __.Context,
    style: Style,
) -> str:
    ''' Produces reStructuredText for exception information.

        Formats exception classes and descriptions in Sphinx-compatible
        reST format. Handles union types of exceptions appropriately.
    '''
    lines: list[ str ] = [ ]
    annotation = information.annotation
    description = _format_description( information.description )
    origin = __.typx.get_origin( annotation )
    if origin in ( __.types.UnionType, __.typx.Union ):
        annotations = __.typx.get_args( annotation )
    else: annotations = ( annotation, )
    for annotation_ in annotations:
        typetext = _format_annotation( annotation_, context, style )
        lines.append(
            f":raises {typetext}: {description}"
            if description else f":raises {typetext}:" )
    return '\n'.join( lines )


def _produce_return_text(
    possessor: __.Documentable,
    information: __.ReturnInformation,
    context: __.Context,
    style: Style,
) -> str:
    ''' Produces reStructuredText for function return information.

        Formats return type and description in Sphinx-compatible reST format.
        Returns empty string for None returns.
    '''
    if information.annotation in ( None, __.types.NoneType ): return ''
    description = _format_description( information.description )
    typetext = _format_annotation( information.annotation, context, style )
    lines: list[ str ] = [ ]
    if description:
        lines.append( f":returns: {description}" )
    lines.append( f":rtype: {typetext}" )
    return '\n'.join( lines )


def _qualify_object_name( # noqa: PLR0911
    objct: object, context: __.Context
) -> str:
    ''' Qualifies an object name for documentation.

        Determines the appropriate fully-qualified name for an object,
        considering builtin types, module namespaces, and qualname attributes.
    '''
    if objct is Ellipsis: return '...'
    if objct is __.types.NoneType: return 'None'
    if objct is __.types.ModuleType: return 'types.ModuleType'
    name = (
        getattr( objct, '__name__', None )
        or _extract_qualident( str( objct ), context ) )
    if name == '<unknown>': return name
    qname = getattr( objct, '__qualname__', None ) or name
    name0 = qname.split( '.', maxsplit = 1 )[ 0 ]
    if name0 in vars( __.builtins ): # int, etc...
        return qname
    if context.invoker_globals and name0 in context.invoker_globals:
        return qname
    mname = getattr( objct, '__module__', None )
    if mname: return f"{mname}.{qname}"
    return name # pragma: no cover


def _stylize_delimiter(
    style: Style,
    delimiters: str,
    content: str,
    prefix: str = '',
) -> str:
    ''' Stylizes delimiters according to the selected style.

        Formats delimiters around content based on the style setting,
        with options for more legible spacing or compact PEP 8 formatting.
    '''
    ld = delimiters[ 0 ]
    rd = delimiters[ 1 ]
    match style:
        case Style.Legible: return f"{prefix}{ld} {content} {rd}"
        case Style.Pep8: return f"{prefix}{ld}{content}{rd}"



# vim: set filetype=python fileencoding=utf-8:
# -*- coding: utf-8 -*-

#============================================================================#
#                                                                            #
#  Licensed under the Apache License, Version 2.0 (the "License");           #
#  you may not use this file except in compliance with the License.          #
#  You may obtain a copy of the License at                                   #
#                                                                            #
#      http://www.apache.org/licenses/LICENSE-2.0                            #
#                                                                            #
#  Unless required by applicable law or agreed to in writing, software       #
#  distributed under the License is distributed on an "AS IS" BASIS,         #
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  #
#  See the License for the specific language governing permissions and       #
#  limitations under the License.                                            #
#                                                                            #
#============================================================================#


''' Interface for extension development. '''

# ruff: noqa: F403,F405


from . import __

from .context import *
from .interfaces import *
from .introspection import *
from .nomina import *


FragmentRectifierArgument: __.typx.TypeAlias = __.typx.Annotated[
    FragmentRectifier, Fname( 'fragment rectifier' ) ]
FragmentsArgumentMultivalent: __.typx.TypeAlias = __.typx.Annotated[
    Fragment,
    Doc(
        ''' Fragments from which to produce a docstring.

            If fragment is a string, then it will be used as an index
            into a table of docstring fragments.
            If fragment is a :pep:`727` ``Doc`` object, then the value of its
            ``documentation`` attribute will be incorporated.
        ''' ),
]
FragmentsNameArgument: __.typx.TypeAlias = __.typx.Annotated[
    str, Fname( 'fragments name' ) ]
IntrospectionLimitNameArgument: __.typx.TypeAlias = __.typx.Annotated[
    str, Fname( 'introspection limit name' ) ]
InvokerGlobalsArgument: __.typx.TypeAlias = __.typx.Annotated[
    __.typx.Optional[ Variables ], Fname( 'invoker globals' ) ]
NotifierArgument: __.typx.TypeAlias = __.typx.Annotated[
    Notifier, Fname( 'notifier' ) ]
PreserveArgument: __.typx.TypeAlias = __.typx.Annotated[
    bool, Doc( ''' Preserve extant docstring? ''' ) ]
ResolverGlobalsArgument: __.typx.TypeAlias = __.typx.Annotated[
    __.typx.Optional[ Variables ], Fname( 'resolver globals' ) ]
ResolverLocalsArgument: __.typx.TypeAlias = __.typx.Annotated[
    __.typx.Optional[ Variables ], Fname( 'resolver locals' ) ]
VisibilityDeciderArgument: __.typx.TypeAlias = __.typx.Annotated[
    VisibilityDecider, Fname( 'visibility decider' ) ]


RendererReturnValue: __.typx.TypeAlias = __.typx.Annotated[
    str, Doc( ''' Rendered docstring fragment. ''' ) ]

class Renderer( __.typx.Protocol ):
    ''' Produces docstring fragment from object and information about it. '''

    @staticmethod
    def __call__(
        possessor: PossessorArgument,
        informations: InformationsArgument,
        context: ContextArgument,
    ) -> RendererReturnValue:
        ''' (Signature for fragment renderer.) '''
        raise NotImplementedError # pragma: no cover

RendererArgument: __.typx.TypeAlias = __.typx.Annotated[
    Renderer, Fname( 'renderer' ) ]



# vim: set filetype=python fileencoding=utf-8:
# -*- coding: utf-8 -*-

#============================================================================#
#                                                                            #
#  Licensed under the Apache License, Version 2.0 (the "License");           #
#  you may not use this file except in compliance with the License.          #
#  You may obtain a copy of the License at                                   #
#                                                                            #
#      http://www.apache.org/licenses/LICENSE-2.0                            #
#                                                                            #
#  Unless required by applicable law or agreed to in writing, software       #
#  distributed under the License is distributed on an "AS IS" BASIS,         #
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  #
#  See the License for the specific language governing permissions and       #
#  limitations under the License.                                            #
#                                                                            #
#============================================================================#


''' Introspection of argument, attribute, and return annotations. '''


from . import __
from . import context as _context
from . import interfaces as _interfaces
from . import nomina as _nomina


_default_default = _interfaces.Default( )
_default_suppress = _interfaces.Default(
    mode = _interfaces.ValuationModes.Suppress )


IntrospectIntrospectionArgument: __.typx.TypeAlias = __.typx.Annotated[
    _context.IntrospectionControl,
    _interfaces.Doc(
        ''' Control settings for introspection behavior. ''' ),
]


def introspect(
    possessor: _interfaces.PossessorArgument, /,
    context: _context.ContextArgument,
    introspection: _context.IntrospectionArgument,
    cache: _interfaces.AnnotationsCacheArgument,
    table: _interfaces.FragmentsTableArgument,
) -> __.cabc.Sequence[ _interfaces.InformationBase ]:
    ''' Introspects object to extract documentable information.

        Dispatches to appropriate introspection function based on the type
        of the object being introspected (class, function, or module).
    '''
    if __.inspect.isclass( possessor ):
        return _introspect_class(
            possessor, context, introspection, cache, table )
    if __.inspect.isfunction( possessor ) and possessor.__name__ != '<lambda>':
        return _introspect_function(
            possessor, context, cache, table )
    if __.inspect.ismodule( possessor ):
        return _introspect_module(
            possessor, context, introspection, cache, table )
    return ( )


def introspect_special_classes( # noqa: PLR0913
    possessor: _interfaces.PossessorClassArgument, /,
    context: _context.ContextArgument,
    introspection: _context.IntrospectionArgument,
    annotations: _interfaces.AnnotationsArgument,
    cache: _interfaces.AnnotationsCacheArgument,
    table: _interfaces.FragmentsTableArgument,
) -> __.typx.Optional[ _interfaces.Informations ]:
    ''' Introspects special classes in Python standard library.

        E.g., enum members are collected as class variables.
    '''
    informations: list[ _interfaces.InformationBase ] = [ ]
    if isinstance( possessor, __.enum.EnumMeta ):
        informations.extend(
            _interfaces.AttributeInformation(
                name = name,
                annotation = possessor,
                description = None,
                association = _interfaces.AttributeAssociations.Class,
                default = _default_suppress )
            for name in possessor.__members__ )
        return informations
    return None


def is_attribute_visible(
    possessor: _interfaces.PossessorArgument,
    name: str,
    annotation: __.typx.Any,
    description: __.typx.Optional[ str ],
) -> bool:
    ''' Determines if attribute should be visible in documentation.

        Default visibility predicate that considers attribute with
        description or public name (not starting with underscore) as visible.

        If attribute possessor is module, then ``__all__`` is considered,
        if it exists.
    '''
    if __.inspect.ismodule( possessor ):
        publics = getattr( possessor, '__all__', None )
        if publics is not None: return name in publics
    return bool( description ) or not name.startswith( '_' )


def reduce_annotation(
    annotation: __.typx.Any,
    context: _context.Context,
    adjuncts: _interfaces.AdjunctsData,
    cache: _interfaces.AnnotationsCache,
) -> __.typx.Any:
    ''' Reduces a complex type annotation to a simpler form.

        Processes type annotations, extracting metadata from Annotated types
        and simplifying complex generic types. Uses cache to avoid redundant
        processing and prevent infinite recursion from reference cycles.
    '''
    annotation_r = cache.access( annotation )
    # Avoid infinite recursion from reference cycles.
    if annotation_r is _interfaces.incomplete:
        emessage = (
            f"Annotation with circular reference {annotation!r}; "
            "returning Any." )
        context.notifier( 'admonition', emessage )
        return cache.enter( annotation, __.typx.Any )
    # TODO: Short-circuit on cache hit.
    #       Need to ensure copy of adjuncts data is retrieved too.
    # if annotation_r is not _interfaces.absent: return annotation_r
    cache.enter( annotation ) # mark as incomplete
    return cache.enter(
        annotation,
        _reduce_annotation_core( annotation, context, adjuncts, cache ) )


def _access_annotations(
    possessor: _nomina.Documentable, /, context: _context.Context
) -> __.cabc.Mapping[ str, __.typx.Any ]:
    ''' Accesses annotations from a documentable object.

        Retrieves annotations with appropriate resolver settings from the
        context. Handles errors gracefully.
    '''
    nomargs: _nomina.Variables = dict( eval_str = True )
    nomargs[ 'globals' ] = context.resolver_globals
    nomargs[ 'locals' ] = context.resolver_locals
    try:
        return __.types.MappingProxyType(
            __.inspect.get_annotations( possessor, **nomargs ) )
    except TypeError as exc:
        emessage = f"Cannot access annotations for {possessor!r}: {exc}"
        context.notifier( 'error', emessage )
        return __.dictproxy_empty


def _classes_sequence_to_union(
    annotation: type | __.cabc.Sequence[ type ]
) -> __.typx.Any:
    ''' Converts a sequence of exception classes to a Union type.

        Used for Raises annotations to convert a sequence of exception
        classes into a Union type for documentation.
    '''
    if not isinstance( annotation, __.cabc.Sequence ):
        return annotation
    return __.funct.reduce( __.operator.or_, annotation )


def _compile_description(
    context: _context.Context,
    adjuncts: _interfaces.AdjunctsData,
    table: _nomina.FragmentsTable,
) -> str:
    ''' Compiles a description from adjuncts data.

        Processes Doc objects and Findex references in adjuncts data
        to create a combined description string with proper formatting.
    '''
    fragments: list[ str ] = [ ]
    for extra in adjuncts.extras:
        if isinstance( extra, _interfaces.Doc ):
            fragments.append( extra.documentation )
        elif isinstance( extra, _interfaces.Fname ):
            name = extra.name
            if name not in table:
                emessage = f"Fragment '{name}' not in provided table."
                context.notifier( 'error', emessage )
            else: fragments.append( table[ name ] )
    return '\n\n'.join(
        context.fragment_rectifier(
            fragment, source = _interfaces.FragmentSources.Annotation )
        for fragment in fragments )


def _determine_default_valuator(
    context: _context.Context,
    adjuncts: _interfaces.AdjunctsData,
) -> _interfaces.Default:
    ''' Determines how default values should be handled.

        Extracts the Default object from adjuncts data or falls back
        to the default Default settings.
    '''
    return next(
        (   extra for extra in adjuncts.extras
            if isinstance( extra, _interfaces.Default ) ),
        _default_default )


def _filter_reconstitute_annotation(
    origin: __.typx.Any,
    arguments: __.cabc.Sequence[ __.typx.Any ],
    context: _context.Context,
    adjuncts: _interfaces.AdjunctsData,
    cache: _interfaces.AnnotationsCache,
) -> __.typx.Any:
    ''' Filters and reconstitutes a generic type annotation.

        After reducing the arguments of a generic type, this function
        reconstitutes the type with the reduced arguments, potentially
        applying transformations based on context.

        Note that any type-adjacent information on arguments is not propagated
        upwards, due to ambiguity in its insertion order relative to
        type-adjacent information on the annotation origin.
    '''
    adjuncts.traits.add( origin.__name__ )
    arguments_r: list[ __.typx.Any ] = [ ]
    adjuncts_ = _interfaces.AdjunctsData( )
    adjuncts_.traits.add( origin.__name__ )
    match len( arguments ):
        case 1:
            arguments_r.append( reduce_annotation(
                arguments[ 0 ], context, adjuncts_, cache ) )
        case _:
            arguments_r.extend( _reduce_annotation_arguments(
                origin, arguments, context, adjuncts_, cache ) )
    # TODO: Apply filters from context, replacing origin as necessary.
    #       E.g., ClassVar -> Union
    #       (Union with one argument returns the argument.)
    try:
        if origin in ( __.types.UnionType, __.typx.Union ):
            # Unions cannot be reconstructed from sequences.
            # TODO: Python 3.11: Unpack into subscript.
            annotation = __.funct.reduce( __.operator.or_, arguments_r )
        else:
            match len( arguments_r ):
                case 1: annotation = origin[ arguments_r[ 0 ] ]
                case _: annotation = origin[ tuple( arguments_r ) ]
    except TypeError as exc:
        emessage = (
            f"Cannot reconstruct {origin.__name__!r} "
            f"with reduced annotations for arguments. Reason: {exc}" )
        context.notifier( 'error', emessage )
        return origin
    return annotation


def _introspect_class(
    possessor: type, /,
    context: _context.Context,
    introspection: _context.IntrospectionControl,
    cache: _interfaces.AnnotationsCache,
    table: _nomina.FragmentsTable,
) -> __.cabc.Sequence[ _interfaces.InformationBase ]:
    ''' Introspects a class to extract documentable information.

        Gathers information about class annotations, potentially considering
        inherited annotations based on introspection control settings. Tries
        special class introspectors first, then falls back to standard
        introspection.
    '''
    annotations_: dict[ str, __.typx.Any ] = { }
    if introspection.class_control.inheritance:
        # Descendant annotations override ancestor annotations.
        for class_ in reversed( possessor.__mro__ ):
            annotations_b = _access_annotations( class_, context )
            annotations_.update( annotations_b )
        annotations = annotations_
    else: annotations = _access_annotations( possessor, context )
    informations: list[ _interfaces.InformationBase ] = [ ]
    for introspector in introspection.class_control.introspectors:
        informations_ = introspector(
            possessor,
            context = context, introspection = introspection,
            annotations = annotations, cache = cache, table = table )
        if informations_ is not None:
            informations.extend( informations_ )
            break
    else:
        informations.extend( _introspect_class_annotations(
            possessor, context, annotations, cache, table ) )
        if introspection.class_control.scan_attributes:
            informations.extend( _introspect_class_attributes(
                possessor, context, annotations ) )
    return tuple( informations )


def _introspect_class_annotations(
    possessor: type, /,
    context: _context.Context,
    annotations: __.cabc.Mapping[ str, __.typx.Any ],
    cache: _interfaces.AnnotationsCache,
    table: _nomina.FragmentsTable,
) -> __.cabc.Sequence[ _interfaces.InformationBase ]:
    ''' Introspects annotations of a class.

        Processes class annotations to extract information about class
        attributes, including their types, descriptions from Doc objects,
        and whether they are class or instance variables.
    '''
    informations: list[ _interfaces.InformationBase ] = [ ]
    for name, annotation in annotations.items( ):
        adjuncts = _interfaces.AdjunctsData( )
        annotation_ = reduce_annotation(
            annotation, context, adjuncts, cache )
        description = _compile_description( context, adjuncts, table )
        if not _is_attribute_visible(
            possessor, name, annotation_, context, adjuncts, description
        ): continue
        association = (
            _interfaces.AttributeAssociations.Class
            if 'ClassVar' in adjuncts.traits
            else _interfaces.AttributeAssociations.Instance )
        default = _determine_default_valuator( context, adjuncts )
        informations.append( _interfaces.AttributeInformation(
            name = name,
            annotation = annotation_,
            description = description,
            association = association,
            default = default ) )
    return informations


def _introspect_class_attributes(
    possessor: type, /,
    context: _context.Context,
    annotations: __.cabc.Mapping[ str, __.typx.Any ],
) -> __.cabc.Sequence[ _interfaces.InformationBase ]:
    ''' Introspects attributes of a class not covered by annotations.

        Examines class attributes that do not have corresponding annotations
        and creates attribute information for those that should be visible.
    '''
    informations: list[ _interfaces.InformationBase ] = [ ]
    adjuncts = _interfaces.AdjunctsData( ) # dummy value
    for name, attribute in __.inspect.getmembers( possessor ):
        if name in annotations: continue # already processed
        if not _is_attribute_visible(
            possessor, name, _interfaces.absent, context, adjuncts, None
        ): continue
        if callable( attribute ): continue # separately documented
        informations.append( _interfaces.AttributeInformation(
            name = name,
            annotation = _interfaces.absent,
            description = None,
            association = _interfaces.AttributeAssociations.Class,
            default = _default_default ) )
    return informations


def _introspect_function(
    possessor: __.cabc.Callable[ ..., __.typx.Any ], /,
    context: _context.Context,
    cache: _interfaces.AnnotationsCache,
    table: _nomina.FragmentsTable,
) -> __.cabc.Sequence[ _interfaces.InformationBase ]:
    ''' Introspects a function to extract documentable information.

        Gathers information about function arguments and return value
        from annotations and signature analysis.
    '''
    annotations = _access_annotations( possessor, context )
    if not annotations: return ( )
    informations: list[ _interfaces.InformationBase ] = [ ]
    try: signature = __.inspect.signature( possessor )
    except ValueError as exc:
        context.notifier(
            'error',
            f"Could not assess signature for {possessor.__qualname__!r}. "
            f"Reason: {exc}" )
        return ( )
    if signature.parameters:
        informations.extend( _introspect_function_valences(
            annotations, signature, context, cache, table ) )
    if 'return' in annotations:
        informations.extend( _introspect_function_return(
            annotations[ 'return' ], context, cache, table ) )
    return tuple( informations )


def _introspect_function_return(
    annotation: __.typx.Any,
    context: _context.Context,
    cache: _interfaces.AnnotationsCache,
    table: _nomina.FragmentsTable,
) -> __.cabc.Sequence[ _interfaces.InformationBase ]:
    ''' Introspects function return annotation.

        Processes function return annotation to extract return type information
        and possible exception information from Raises annotations.
    '''
    informations: list[ _interfaces.InformationBase ] = [ ]
    adjuncts = _interfaces.AdjunctsData( )
    annotation_ = reduce_annotation( annotation, context, adjuncts, cache )
    description = _compile_description( context, adjuncts, table )
    informations.append(
        _interfaces.ReturnInformation(
            annotation = annotation_, description = description ) )
    informations.extend(
        _interfaces.ExceptionInformation(
            annotation = _classes_sequence_to_union( extra.classes ),
            description = extra.description )
        for extra in adjuncts.extras
        if isinstance( extra, _interfaces.Raises ) )
    return tuple( informations )


def _introspect_function_valences(
    annotations: __.cabc.Mapping[ str, __.typx.Any ],
    signature: __.inspect.Signature,
    context: _context.Context,
    cache: _interfaces.AnnotationsCache,
    table: _nomina.FragmentsTable,
) -> __.cabc.Sequence[ _interfaces.ArgumentInformation ]:
    ''' Introspects function parameters to extract argument information.

        Processes function signature and annotations to create information
        about function arguments, including their types, descriptions, and
        default value handling.
    '''
    informations: list[ _interfaces.ArgumentInformation ] = [ ]
    for name, param in signature.parameters.items( ):
        annotation = annotations.get( name, param.annotation )
        adjuncts = _interfaces.AdjunctsData( )
        if annotation is param.empty:
            annotation_ = _interfaces.absent
            description = None
        else:
            annotation_ = reduce_annotation(
                annotation, context, adjuncts, cache )
            description = _compile_description( context, adjuncts, table )
        if param.default is param.empty: default = _default_suppress
        else: default = _determine_default_valuator( context, adjuncts )
        informations.append( _interfaces.ArgumentInformation(
            name = name,
            annotation = annotation_,
            description = description,
            paramspec = param,
            default = default ) )
    return tuple( informations )


def _introspect_module(
    possessor: __.types.ModuleType, /,
    context: _context.Context,
    introspection: _context.IntrospectionControl,
    cache: _interfaces.AnnotationsCache,
    table: _nomina.FragmentsTable,
) -> __.cabc.Sequence[ _interfaces.InformationBase ]:
    ''' Introspects a module to extract documentable information.

        Gathers information about module annotations and potentially about
        module attributes based on introspection control settings.
    '''
    annotations = _access_annotations( possessor, context )
    if not annotations: return ( )
    informations: list[ _interfaces.InformationBase ] = [ ]
    informations.extend( _introspect_module_annotations(
        possessor, context, annotations, cache, table ) )
    if introspection.module_control.scan_attributes:
        informations.extend( _introspect_module_attributes(
            possessor, context, annotations ) )
    return tuple( informations )


def _introspect_module_annotations(
    possessor: __.types.ModuleType, /,
    context: _context.Context,
    annotations: __.cabc.Mapping[ str, __.typx.Any ],
    cache: _interfaces.AnnotationsCache,
    table: _nomina.FragmentsTable,
) -> __.cabc.Sequence[ _interfaces.InformationBase ]:
    ''' Introspects annotations of a module.

        Processes module annotations to extract information about module
        attributes, including their types and descriptions from Doc objects.
    '''
    informations: list[ _interfaces.InformationBase ] = [ ]
    for name, annotation in annotations.items( ):
        adjuncts = _interfaces.AdjunctsData( )
        annotation_ = reduce_annotation(
            annotation, context, adjuncts, cache )
        description = _compile_description( context, adjuncts, table )
        if not _is_attribute_visible(
            possessor, name, annotation_, context, adjuncts, description
        ): continue
        default = _determine_default_valuator( context, adjuncts )
        informations.append( _interfaces.AttributeInformation(
            name = name,
            annotation = annotation_,
            description = description,
            association = _interfaces.AttributeAssociations.Module,
            default = default ) )
    return informations


def _introspect_module_attributes(
    possessor: __.types.ModuleType, /,
    context: _context.Context,
    annotations: __.cabc.Mapping[ str, __.typx.Any ],
) -> __.cabc.Sequence[ _interfaces.InformationBase ]:
    ''' Introspects attributes of a module not covered by annotations.

        Examines module attributes that do not have corresponding annotations
        and creates attribute information for those that should be visible.
    '''
    informations: list[ _interfaces.InformationBase ] = [ ]
    adjuncts = _interfaces.AdjunctsData( ) # dummy value
    attribute: object
    for name, attribute in __.inspect.getmembers( possessor ):
        if name in annotations: continue # already processed
        if not _is_attribute_visible(
            possessor, name, _interfaces.absent, context, adjuncts, None
        ): continue
        if callable( attribute ): continue # separately documented
        informations.append( _interfaces.AttributeInformation(
            name = name,
            annotation = _interfaces.absent,
            description = None,
            association = _interfaces.AttributeAssociations.Module,
            default = _default_default ) )
    return informations


def _is_attribute_visible( # noqa: PLR0913
    possessor: _nomina.Documentable,
    name: str,
    annotation: __.typx.Any,
    context: _context.Context,
    adjuncts: _interfaces.AdjunctsData,
    description: __.typx.Optional[ str ],
) -> bool:
    ''' Determines if an attribute should be visible in documentation.

        Checks for explicit visibility settings in adjuncts data and falls
        back to the context's visibility decider if the visibility is set
        to Default.
    '''
    visibility = next(
        (   extra for extra in adjuncts.extras
            if isinstance( extra, _interfaces.Visibilities ) ),
        _interfaces.Visibilities.Default )
    match visibility:
        case _interfaces.Visibilities.Conceal: return False
        case _interfaces.Visibilities.Reveal: return True
        case _:
            return context.visibility_decider(
                possessor, name, annotation, description )


def _reduce_annotation_arguments(
    origin: __.typx.Any,
    arguments: __.cabc.Sequence[ __.typx.Any ],
    context: _context.Context,
    adjuncts: _interfaces.AdjunctsData,
    cache: _interfaces.AnnotationsCache,
) -> __.cabc.Sequence[ __.typx.Any ]:
    ''' Reduces the arguments of a generic type annotation.

        Processes the arguments of a generic type like List[T] or Dict[K, V]
        and returns the reduced forms of those arguments. Special handling
        for Callable types.
    '''
    if __.inspect.isclass( origin ) and issubclass( origin, __.cabc.Callable ):
        return _reduce_annotation_for_callable(
            arguments, context, adjuncts.copy( ), cache )
    return tuple(
        reduce_annotation( argument, context, adjuncts.copy( ), cache )
        for argument in arguments )


def _reduce_annotation_core(
    annotation: __.typx.Any,
    context: _context.Context,
    adjuncts: _interfaces.AdjunctsData,
    cache: _interfaces.AnnotationsCache,
) -> __.typx.Any:
    ''' Core implementation of annotation reduction.

        Handles the reduction of complex type annotations into simpler forms,
        extracting metadata from Annotated types and processing generic types.
        Returns the reduced annotation.
    '''
    origin = __.typx.get_origin( annotation )
    # bare types, Ellipsis, typing.Any, typing.LiteralString, typing.Never,
    # typing.TypeVar have no origin; taken as-is
    # typing.Literal is considered fully reduced; taken as-is
    if origin in ( None, __.typx.Literal ): return annotation
    arguments = __.typx.get_args( annotation )
    if not arguments: return annotation
    if origin is __.typx.Annotated:
        adjuncts.extras.extend( arguments[ 1 : ] )
        return reduce_annotation(
            annotation.__origin__, context, adjuncts, cache )
    return _filter_reconstitute_annotation(
        origin, arguments, context, adjuncts, cache )


def _reduce_annotation_for_callable(
    arguments: __.cabc.Sequence[ __.typx.Any ],
    context: _context.Context,
    adjuncts: _interfaces.AdjunctsData,
    cache: _interfaces.AnnotationsCache,
) -> tuple[ list[ __.typx.Any ] | __.types.EllipsisType, __.typx.Any ]:
    ''' Reduces annotations for Callable types.

        Special handling for Callable type annotations, which have a tuple
        of (arguments, return_type). Processes the arguments list and return
        type separately and returns the reduced forms.
    '''
    farguments, freturn = arguments
    if farguments is Ellipsis:
        farguments_r = Ellipsis
    else:
        farguments_r = [
            reduce_annotation( element, context, adjuncts.copy( ), cache )
            for element in farguments ]
    freturn_r = (
        reduce_annotation( freturn, context, adjuncts.copy( ), cache ) )
    return ( farguments_r, freturn_r )



# vim: set filetype=python fileencoding=utf-8:
# -*- coding: utf-8 -*-

#============================================================================#
#                                                                            #
#  Licensed under the Apache License, Version 2.0 (the "License");           #
#  you may not use this file except in compliance with the License.          #
#  You may obtain a copy of the License at                                   #
#                                                                            #
#      http://www.apache.org/licenses/LICENSE-2.0                            #
#                                                                            #
#  Unless required by applicable law or agreed to in writing, software       #
#  distributed under the License is distributed on an "AS IS" BASIS,         #
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  #
#  See the License for the specific language governing permissions and       #
#  limitations under the License.                                            #
#                                                                            #
#============================================================================#


''' Docstring assembly and decoration. '''
# TODO? with_docstring_defer
#       Registers with_docstring partial function in registry.
#       Registry can be executed after modules are loaded and all string
#       annotations should be resolvable.


from . import __
from . import factories as _factories
from . import renderers as _renderers
from . import xtnsapi as _xtnsapi


_visitees: __.weakref.WeakSet[ _xtnsapi.Documentable ] = __.weakref.WeakSet( )


context_default: __.typx.Annotated[
    _xtnsapi.Context,
    _xtnsapi.Doc(
        ''' Default context for introspection and rendering. ''' ),
    _xtnsapi.Fname( 'context' ),
    _xtnsapi.Default( mode = _xtnsapi.ValuationModes.Suppress ),
] = _factories.produce_context( )
introspection_default: __.typx.Annotated[
    _xtnsapi.IntrospectionControl,
    _xtnsapi.Doc( ''' Default introspection control. ''' ),
    _xtnsapi.Fname( 'introspection' ),
    _xtnsapi.Default( mode = _xtnsapi.ValuationModes.Suppress ),
] = _xtnsapi.IntrospectionControl( )
renderer_default: __.typx.Annotated[
    _xtnsapi.Renderer,
    _xtnsapi.Doc( ''' Default renderer for docstring fragments. ''' ),
    _xtnsapi.Fname( 'renderer' ),
    _xtnsapi.Default( mode = _xtnsapi.ValuationModes.Suppress ),
] = _renderers.sphinxad.produce_fragment


def assign_module_docstring( # noqa: PLR0913
    module: _xtnsapi.Module, /,
    *fragments: _xtnsapi.FragmentsArgumentMultivalent,
    context: _xtnsapi.ContextArgument = context_default,
    introspection: _xtnsapi.IntrospectionArgument = introspection_default,
    preserve: _xtnsapi.PreserveArgument = True,
    renderer: _xtnsapi.RendererArgument = renderer_default,
    table: _xtnsapi.FragmentsTableArgument = __.dictproxy_empty,
) -> None:
    ''' Assembles docstring from fragments and assigns it to module. '''
    if isinstance( module, str ):
        module = __.sys.modules[ module ]
    _decorate(
        module,
        context = context,
        introspection = introspection,
        preserve = preserve,
        renderer = renderer,
        fragments = fragments,
        table = table )


def with_docstring(
    *fragments: _xtnsapi.FragmentsArgumentMultivalent,
    context: _xtnsapi.ContextArgument = context_default,
    introspection: _xtnsapi.IntrospectionArgument = introspection_default,
    preserve: _xtnsapi.PreserveArgument = True,
    renderer: _xtnsapi.RendererArgument = renderer_default,
    table: _xtnsapi.FragmentsTableArgument = __.dictproxy_empty,
) -> _xtnsapi.Decorator[ _xtnsapi.D ]:
    ''' Assembles docstring from fragments and decorates object with it. '''
    def decorate( objct: _xtnsapi.D ) -> _xtnsapi.D:
        _decorate(
            objct,
            context = context,
            introspection = introspection,
            preserve = preserve,
            renderer = renderer,
            fragments = fragments,
            table = table )
        return objct

    return decorate


def _check_module_recursion(
    objct: object, /,
    introspection: _xtnsapi.IntrospectionControl,
    mname: str
) -> __.typx.TypeIs[ __.types.ModuleType ]:
    ''' Checks if a module should be recursively documented.

        Returns True if the object is a module that should be recursively
        documented based on the introspection control and module name prefix.
    '''
    if (    introspection.targets & _xtnsapi.IntrospectionTargets.Module
        and __.inspect.ismodule( objct )
    ): return objct.__name__.startswith( f"{mname}." )
    return False


def _collect_fragments(
    objct: _xtnsapi.Documentable, /, context: _xtnsapi.Context, fqname: str
) -> _xtnsapi.Fragments:
    ''' Collects docstring fragments from an object.

        Retrieves the sequence of fragments stored on the object using the
        fragments_name from the context. Validates that the fragments are
        of the expected types.
    '''
    fragments: _xtnsapi.Fragments = (
        # Fragments can come from base class or metaclass.
        # We only care about fragments on class itself.
        objct.__dict__.get( context.fragments_name, ( ) )
        if __.inspect.isclass( objct )
        else getattr( objct, context.fragments_name, ( ) ) )
    if (    isinstance( fragments, ( bytes, str ) )
        or not isinstance( fragments, __.cabc.Sequence )
    ):
        emessage = f"Invalid fragments sequence on {fqname}: {fragments!r}"
        context.notifier( 'error', emessage )
        fragments = ( )
    for fragment in fragments:
        if not isinstance( fragment, ( str, _xtnsapi.Doc ) ):
            emessage = f"Invalid fragment on {fqname}: {fragment!r}"
            context.notifier( 'error', emessage )
    return fragments


def _consider_class_attribute( # noqa: C901,PLR0913
    attribute: object, /,
    context: _xtnsapi.Context,
    introspection: _xtnsapi.IntrospectionControl,
    pmname: str, pqname: str, aname: str,
) -> tuple[ __.typx.Optional[ _xtnsapi.Documentable ], bool ]:
    ''' Considers whether a class attribute should be documented.

        Examines a class attribute to determine if it should be included
        in the documentation process based on introspection targets and
        class ownership. Returns the documentable attribute and a flag
        indicating whether the surface attribute needs updating.
    '''
    if _check_module_recursion( attribute, introspection, pmname ):
        return attribute, False
    attribute_ = None
    update_surface = False
    if (    not attribute_
        and introspection.targets & _xtnsapi.IntrospectionTargets.Class
        and __.inspect.isclass( attribute )
    ): attribute_ = attribute
    if (    not attribute_
        and introspection.targets & _xtnsapi.IntrospectionTargets.Descriptor
    ):
        if isinstance( attribute, property ) and attribute.fget:
            # Examine docstring and signature of getter method on property.
            attribute_ = attribute.fget
            update_surface = True
        # TODO: Apply custom processors from context.
        elif __.inspect.isdatadescriptor( attribute ):
            # Ignore descriptors which we do not know how to handle.
            return None, False
    if (    not attribute_
        and introspection.targets & _xtnsapi.IntrospectionTargets.Function
    ):
        if __.inspect.ismethod( attribute ):
            # Methods proxy docstrings from their core functions.
            attribute_ = attribute.__func__
        elif __.inspect.isfunction( attribute ) and aname != '<lambda>':
            attribute_ = attribute
    if attribute_:
        mname = getattr( attribute_, '__module__', None )
        if not mname or mname != pmname:
            attribute_ = None
    if attribute_:
        qname = getattr( attribute_, '__qualname__', None )
        if not qname or not qname.startswith( f"{pqname}." ):
            attribute_ = None
    return attribute_, update_surface


def _consider_module_attribute(
    attribute: object, /,
    context: _xtnsapi.Context,
    introspection: _xtnsapi.IntrospectionControl,
    pmname: str, aname: str,
) -> tuple[ __.typx.Optional[ _xtnsapi.Documentable ], bool ]:
    ''' Considers whether a module attribute should be documented.

        Examines a module attribute to determine if it should be included
        in the documentation process based on introspection targets and
        module ownership. Returns the documentable attribute and a flag
        indicating whether the surface attribute needs updating.
    '''
    if _check_module_recursion( attribute, introspection, pmname ):
        return attribute, False
    attribute_ = None
    update_surface = False
    if (    not attribute_
        and introspection.targets & _xtnsapi.IntrospectionTargets.Class
        and __.inspect.isclass( attribute )
    ): attribute_ = attribute
    if (    not attribute_
        and introspection.targets & _xtnsapi.IntrospectionTargets.Function
        and __.inspect.isfunction( attribute ) and aname != '<lambda>'
    ): attribute_ = attribute
    if attribute_:
        mname = getattr( attribute_, '__module__', None )
        if not mname or mname != pmname:
            attribute_ = None
    return attribute_, update_surface


def _decorate( # noqa: PLR0913
    objct: _xtnsapi.Documentable, /,
    context: _xtnsapi.Context,
    introspection: _xtnsapi.IntrospectionControl,
    preserve: bool,
    renderer: _xtnsapi.Renderer,
    fragments: _xtnsapi.Fragments,
    table: _xtnsapi.FragmentsTable,
) -> None:
    ''' Decorates an object with assembled docstring.

        Handles core docstring decoration and potentially recursive decoration
        of the object's attributes based on introspection control settings.
        Prevents multiple decoration of the same object.
    '''
    if objct in _visitees: return # Prevent multiple decoration.
    _visitees.add( objct )
    if introspection.targets:
        if __.inspect.isclass( objct ):
            _decorate_class_attributes(
                objct,
                context = context,
                introspection = introspection,
                preserve = preserve,
                renderer = renderer,
                table = table )
        elif __.inspect.ismodule( objct ):
            _decorate_module_attributes(
                objct,
                context = context,
                introspection = introspection,
                preserve = preserve,
                renderer = renderer,
                table = table )
    if __.inspect.ismodule( objct ): fqname = objct.__name__
    else: fqname = f"{objct.__module__}.{objct.__qualname__}"
    fragments_ = _collect_fragments( objct, context, fqname )
    if not fragments_: fragments_ = fragments
    _decorate_core(
        objct,
        context = context,
        introspection = introspection,
        preserve = preserve,
        renderer = renderer,
        fragments = fragments_,
        table = table )


def _decorate_core( # noqa: PLR0913
    objct: _xtnsapi.Documentable, /,
    context: _xtnsapi.Context,
    introspection: _xtnsapi.IntrospectionControl,
    preserve: bool,
    renderer: _xtnsapi.Renderer,
    fragments: _xtnsapi.Fragments,
    table: _xtnsapi.FragmentsTable,
) -> None:
    ''' Core implementation of docstring decoration.

        Assembles a docstring from fragments, existing docstring (if
        preserved), and introspection results. Assigns the assembled docstring
        to the object.
    '''
    fragments_: list[ str ] = [ ]
    if preserve and ( fragment := getattr( objct, '__doc__', None ) ):
        fragments_.append( context.fragment_rectifier(
            fragment, source = _xtnsapi.FragmentSources.Docstring ) )
    fragments_.extend(
        _process_fragments_argument( context, fragments, table ) )
    if introspection.enable:
        cache = _xtnsapi.AnnotationsCache( )
        informations = (
            _xtnsapi.introspect(
                objct,
                context = context, introspection = introspection,
                cache = cache, table = table ) )
        fragments_.append( context.fragment_rectifier(
            renderer( objct, informations, context = context ),
            source = _xtnsapi.FragmentSources.Renderer ) )
    docstring = '\n\n'.join(
        fragment for fragment in filter( None, fragments_ ) ).rstrip( )
    objct.__doc__ = docstring if docstring else None


def _decorate_class_attributes( # noqa: PLR0913
    objct: type, /,
    context: _xtnsapi.Context,
    introspection: _xtnsapi.IntrospectionControl,
    preserve: bool,
    renderer: _xtnsapi.Renderer,
    table: _xtnsapi.FragmentsTable,
) -> None:
    ''' Decorates attributes of a class with assembled docstrings.

        Iterates through relevant class attributes, collects fragments,
        and applies appropriate docstring decoration to each attribute.
    '''
    pmname = objct.__module__
    pqname = objct.__qualname__
    for aname, attribute, surface_attribute in (
        _survey_class_attributes( objct, context, introspection )
    ):
        fqname = f"{pmname}.{pqname}.{aname}"
        introspection_ = _limit_introspection(
            attribute, context, introspection, fqname )
        introspection_ = introspection_.evaluate_limits_for( attribute )
        if not introspection_.enable: continue
        _decorate(
            attribute,
            context = context,
            introspection = introspection_,
            preserve = preserve,
            renderer = renderer,
            fragments = ( ),
            table = table )
        if attribute is not surface_attribute:
            surface_attribute.__doc__ = attribute.__doc__


def _decorate_module_attributes( # noqa: PLR0913
    module: __.types.ModuleType, /,
    context: _xtnsapi.Context,
    introspection: _xtnsapi.IntrospectionControl,
    preserve: bool,
    renderer: _xtnsapi.Renderer,
    table: _xtnsapi.FragmentsTable,
) -> None:
    ''' Decorates attributes of a module with assembled docstrings.

        Iterates through relevant module attributes, collects fragments,
        and applies appropriate docstring decoration to each attribute.
    '''
    pmname = module.__name__
    for aname, attribute, surface_attribute in (
        _survey_module_attributes( module, context, introspection )
    ):
        fqname = f"{pmname}.{aname}"
        introspection_ = _limit_introspection(
            attribute, context, introspection, fqname )
        introspection_ = introspection_.evaluate_limits_for( attribute )
        if not introspection_.enable: continue
        _decorate(
            attribute,
            context = context,
            introspection = introspection_,
            preserve = preserve,
            renderer = renderer,
            fragments = ( ),
            table = table )
        if attribute is not surface_attribute: # pragma: no cover
            surface_attribute.__doc__ = attribute.__doc__


def _limit_introspection(
    objct: _xtnsapi.Documentable, /,
    context: _xtnsapi.Context,
    introspection: _xtnsapi.IntrospectionControl,
    fqname: str,
) -> _xtnsapi.IntrospectionControl:
    ''' Limits introspection based on object-specific constraints.

        Returns a new IntrospectionControl that respects the limits
        specified by the object being documented. This allows objects
        to control how deeply they are introspected.
    '''
    limit: _xtnsapi.IntrospectionLimit = (
        getattr(
            objct,
            context.introspection_limit_name,
            _xtnsapi.IntrospectionLimit( ) ) )
    if not isinstance( limit, _xtnsapi.IntrospectionLimit ):
        emessage = f"Invalid introspection limit on {fqname}: {limit!r}"
        context.notifier( 'error', emessage )
        return introspection
    return introspection.with_limit( limit )


def _process_fragments_argument(
    context: _xtnsapi.Context,
    fragments: _xtnsapi.Fragments,
    table: _xtnsapi.FragmentsTable,
) -> __.cabc.Sequence[ str ]:
    ''' Processes fragments argument into a sequence of string fragments.

        Converts Doc objects to their documentation strings and resolves
        string references to the fragments table. Returns a sequence of
        rectified fragment strings.
    '''
    fragments_: list[ str ] = [ ]
    for fragment in fragments:
        if isinstance( fragment, _xtnsapi.Doc ):
            fragment_r = fragment.documentation
        elif isinstance( fragment, str ):
            if fragment not in table:
                emessage = f"Fragment '{fragment}' not in provided table."
                context.notifier( 'error', emessage )
                continue
            fragment_r = table[ fragment ]
        else:
            emessage = f"Fragment {fragment!r} is invalid. Must be Doc or str."
            context.notifier( 'error', emessage )
            continue
        fragments_.append( context.fragment_rectifier(
            fragment_r, source = _xtnsapi.FragmentSources.Argument ) )
    return fragments_


def _survey_class_attributes(
    possessor: type, /,
    context: _xtnsapi.Context,
    introspection: _xtnsapi.IntrospectionControl,
) -> __.cabc.Iterator[ tuple[ str, _xtnsapi.Documentable, object ] ]:
    ''' Surveys attributes of a class for documentation.

        Yields a sequence of (name, attribute, surface_attribute) tuples
        representing documentable attributes of the class. The surface
        attribute may differ from attribute in cases like properties where the
        attribute's getter method holds the documentation.
    '''
    pmname = possessor.__module__
    pqname = possessor.__qualname__
    for aname, attribute in __.inspect.getmembers( possessor ):
        attribute_, update_surface = (
            _consider_class_attribute(
                attribute, context, introspection, pmname, pqname, aname ) )
        if attribute_ is None: continue
        if update_surface:
            yield aname, attribute_, attribute
            continue
        yield aname, attribute_, attribute_


def _survey_module_attributes(
    possessor: __.types.ModuleType, /,
    context: _xtnsapi.Context,
    introspection: _xtnsapi.IntrospectionControl,
) -> __.cabc.Iterator[ tuple[ str, _xtnsapi.Documentable, object ] ]:
    ''' Surveys attributes of a module for documentation.

        Yields a sequence of (name, attribute, surface_attribute) tuples
        representing documentable attributes of the module. The surface
        attribute may differ from attribute in cases where the actual
        documented object is not directly accessible.
    '''
    pmname = possessor.__name__
    for aname, attribute in __.inspect.getmembers( possessor ):
        attribute_, update_surface = (
            _consider_module_attribute(
                attribute, context, introspection, pmname, aname ) )
        if attribute_ is None: continue
        if update_surface: # pragma: no cover
            yield aname, attribute_, attribute
            continue
        yield aname, attribute_, attribute_



# vim: set filetype=python fileencoding=utf-8:
# -*- coding: utf-8 -*-

#============================================================================#
#                                                                            #
#  Licensed under the Apache License, Version 2.0 (the "License");           #
#  you may not use this file except in compliance with the License.          #
#  You may obtain a copy of the License at                                   #
#                                                                            #
#      http://www.apache.org/licenses/LICENSE-2.0                            #
#                                                                            #
#  Unless required by applicable law or agreed to in writing, software       #
#  distributed under the License is distributed on an "AS IS" BASIS,         #
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  #
#  See the License for the specific language governing permissions and       #
#  limitations under the License.                                            #
#                                                                            #
#============================================================================#


''' Common imports and type aliases used throughout the package. '''

# ruff: noqa: F401


import                      builtins
import collections.abc as   cabc
import dataclasses as       dcls
import                      enum
import functools as         funct
import                      inspect
import itertools as         itert
import                      operator
import                      re
import                      sys
import                      types
import                      warnings
import                      weakref

from pathlib import Path

import typing_extensions as typx
# --- BEGIN: Injected by Copier ---
# --- END: Injected by Copier ---



# vim: set filetype=python fileencoding=utf-8:
# -*- coding: utf-8 -*-

#============================================================================#
#                                                                            #
#  Licensed under the Apache License, Version 2.0 (the "License");           #
#  you may not use this file except in compliance with the License.          #
#  You may obtain a copy of the License at                                   #
#                                                                            #
#      http://www.apache.org/licenses/LICENSE-2.0                            #
#                                                                            #
#  Unless required by applicable law or agreed to in writing, software       #
#  distributed under the License is distributed on an "AS IS" BASIS,         #
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  #
#  See the License for the specific language governing permissions and       #
#  limitations under the License.                                            #
#                                                                            #
#============================================================================#


''' Docstring fragments. '''


from . import imports as __


_FragmentsTable: __.typx.TypeAlias = __.cabc.Mapping[ str, str ]
fragments: _FragmentsTable = __.types.MappingProxyType( {

    'context':
    ''' Data transfer object for various behaviors.

        Controls how annotations are resolved and how fragments are
        processed and rendered.
    ''',

    'fragment rectifier':
    ''' Cleans and normalizes documentation fragment. ''',

    'fragments name':
    ''' Name of class attribute which stores documentation fragments. ''',

    'introspection':
    ''' Controls on introspection behavior.

        Is introspection enabled?
        Which kinds of objects to recursively document?
        Etc...
    ''',

    'introspection limit name':
    ''' Name of class attribute which stores introspection limit. ''',

    'invoker globals':
    ''' Dictionary of globals from the frame of a caller.

        Used by renderers for determing whether to fully-qualify a name.
    ''',

    'notifier': ''' Notifies of warnings and errors. ''',

    'renderer':
    ''' Produces docstring fragment from object and information about it. ''',

    'resolver globals':
    ''' Dictionary of globals for annotation resolution.

        Used for resolving string annotations.
    ''',

    'resolver locals':
    ''' Dictionary of locals for annotation resolution.

        Used for resolving string annotations.
    ''',

    'visibility decider':
    ''' Decides if attribute should have visible documentation. ''',

} )
